
/****************************************************************************************************
curses-plus-utils, aims to provide a C++-like interface to program with curses (ncurses, pdcurses)
library, and provide a set of more complex controls based on the primitive WINDOW* from curses.

Copyright (C) 2008 Andrius da Costa Ribas <andriusmao@gmail.com>,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

 ****************************************************************************************************/

#include <curses.h>
#include <cstdarg>
#include <string>


class COLOR;
class PAIR;
class ATTR;
class WINPLUS;
class WPLUSMAIN;

class COLOR
{
	short color;
protected:
// the following attribute will use ncurses-specific functions
#ifdef NCURSES_VERSION
	static bool uses_default;
#endif
	bool lastoper;
	friend class PAIR;
	friend class WINPLUS;
public:
	COLOR(short col=COLOR_WHITE);
	COLOR& setrgb(short int r, short int g, short int b);
	short get();
	COLOR& getrgb(short int& r, short int& g, short int& b);
	COLOR& operator=(short col);
	bool good();
// the following methods will use ncurses-specific functions
#ifdef NCURSES_VERSION
	static int use_default();
	static int assume_default(int fg, int bg);
	static bool using_default();
#endif
	static bool can_change();
	static bool has();
	static int start();
};
bool COLOR::uses_default=false;

class PAIR
{
	COLOR fore;
	COLOR back;
	short pair;
	static short nextpair;
	friend class COLOR;
	friend class ATTR;
	friend class WINPLUS;
public:
	PAIR(COLOR f, COLOR b);
	PAIR(short f, short b);
	PAIR();
	PAIR& change(COLOR f, COLOR b);
	PAIR& change(short f, short b);
	ATTR& operator+(ATTR a);
};
short PAIR::nextpair=1;
// NOTE: pairs are not removable yet (or maybe they won't be removable)

class ATTR
{
	int attrs;
	friend class WINPLUS;
public:
	ATTR& operator=(long unsigned int a);
	ATTR& operator=(PAIR a);
	ATTR& operator+(long unsigned int a);
	ATTR& operator+(ATTR a);
	ATTR& operator+(PAIR a);
	ATTR(long unsigned int a=A_NORMAL);
	ATTR(PAIR& a);
};



class WINPLUS
{
	WINDOW* W;
protected:
	bool lastoper; //whether the last operation was successful
	int lastscan; // how many variables were mapped in last scanw-like operation
	void setwpointer(WINDOW *reference); /* apenas heranças podem modificar o ponteiro */
	WINDOW *getwpointer(); /* apenas heranças podem acessar o ponteiro */
	WINPLUS(){} /* para herança de WPLUSMAIN*/
public:
	WINPLUS(int nlines, int ncols, int begin_y, int begin_x); //newwin
	WINPLUS(const WINPLUS& win); //dupwin
	~WINPLUS();
	void destroy();
	
	bool good(); // returns status of last operation
	int scanned(); // how many variables were mapped in last scanw-like operation
	
	int chget(); // equiv. wgetch
	WINPLUS& keypad(bool active); // equiv. keypad
	WINPLUS& noecho(bool active); // equiv. noecho
	WINPLUS& move(int y, int x);
	
	
	// wprintw-like / wscanw-like
	WINPLUS& print(char *mask, ...);
	WINPLUS& vprint(char *mask, va_list args);
	WINPLUS& scan(char *mask, ...);
	WINPLUS& vscan(char *mask, va_list args);
	
	
	// cout-like
	WINPLUS& operator<<(int k);
	WINPLUS& operator<<(float k);
	WINPLUS& operator<<(char k);
	WINPLUS& operator<<(double k);
	WINPLUS& operator<<(char* k);
	WINPLUS& operator<<(std::string k);
	
	
	// cin-like
	WINPLUS& operator>>(int& k);
	WINPLUS& operator>>(float& k);
	WINPLUS& operator>>(char& k);
	WINPLUS& operator>>(double& k);
	WINPLUS& operator>>(char*& k);
	WINPLUS& operator>>(std::string& k);
	
	// wattr- functions (here called a- functions), some names can't be repeated, because they're macros, "c_" stands for "wcolor_"
	WINPLUS& aoff(int attrs);
	WINPLUS& aon(int attrs);
	WINPLUS& aset(int attrs);
	WINPLUS& cset(short color_pair_number, void* opts);
	WINPLUS& cset(short color_pair_number);
	WINPLUS& astandend(); // plans to substitute these two for astand(bool standout);
	WINPLUS& astandout();
	WINPLUS& aget(attr_t& attrs, short& pair, void *opts);
	WINPLUS& aget(attr_t& attrs, short& pair);
	WINPLUS& aoff(attr_t attrs, void *opts);
	WINPLUS& aoff(attr_t attrs);
	WINPLUS& aon(attr_t attrs, void *opts);
	WINPLUS& aon(attr_t attrs);
	WINPLUS& aset(attr_t attrs, short pair, void *opts);
	WINPLUS& aset(attr_t attrs, short pair);
	WINPLUS& chga(int n, attr_t attr, short color, const void *opts);
	WINPLUS& chga(int n, attr_t attr, short color);
	WINPLUS& aon(ATTR attrs);
	WINPLUS& aoff(ATTR attrs);
	WINPLUS& aset(ATTR attrs);
	WINPLUS& aoff(PAIR attrs);
	WINPLUS& aon(PAIR attrs);
	WINPLUS& aset(PAIR attrs);
	WINPLUS& aget(attr_t& attrs, PAIR& pair, void *opts=NULL);
	WINPLUS& aset(attr_t attrs, PAIR pair, void *opts=NULL);
	WINPLUS& chga(int n, attr_t attr, COLOR color, const void *opts=NULL);
	
	//refresh
	WINPLUS& fresh();
	WINPLUS& noutfresh();
};

class WPLUSMAIN: public WINPLUS
{
public:
	WPLUSMAIN();
	~WPLUSMAIN();
};





ATTR& ATTR::operator=(long unsigned int a)
{
	attrs=a;
	return *this;
}

ATTR& ATTR::operator=(PAIR a)
{
	attrs=COLOR_PAIR(a.pair);
	return *this;
}

ATTR& ATTR::operator+(long unsigned int a)
{
	attrs=attrs | a;
	return *this;
}

ATTR& ATTR::operator+(ATTR a)
{
	this->attrs=this->attrs | a.attrs;
	return *this;
}

ATTR& ATTR::operator+(PAIR a)
{
	attrs=attrs | COLOR_PAIR(a.pair);
	return *this;
}

ATTR::ATTR(long unsigned int a): attrs(a) {}

ATTR::ATTR(PAIR& a)
{
	attrs=COLOR_PAIR(a.pair);
}






ATTR& PAIR::operator+(ATTR a)
{
	return a+(*this);
}

PAIR& PAIR::change(COLOR f, COLOR b)
{
	fore=f;
	back=b;
	init_pair(pair,f.color,b.color);
	return *this;
}

PAIR& PAIR::change(short f, short b)
{
	fore=f;
	back=b;
	init_pair(pair,f,b);
	return *this;
}

PAIR::PAIR(COLOR f, COLOR b)
{
	fore=f;
	back=b;
	init_pair(nextpair,f.color,b.color);
	pair=nextpair;
	nextpair++;
	
}

PAIR::PAIR(short f, short b)
{
	fore=f;
	back=b;
	init_pair(nextpair,f,b);
	pair=nextpair;
	nextpair++;
}

PAIR::PAIR()
{
#ifdef NCURSES_VERSION
	if (COLOR::using_default())
	{
		fore=-1;
		back=-1;
		pair=0;
	}
	else
	{
#endif
		fore=COLOR_WHITE;
		fore=COLOR_BLACK;
		pair=0;
#ifdef NCURSES_VERSION
	}
#endif
}














COLOR::COLOR(short col)
{
	// CAN BE IMPROVED
	if(col>COLORS || col<(-1))
	{
		lastoper=false;
		color=COLOR_WHITE;
	}
	else
	{
		lastoper=true;
		color=col;
	}
}


COLOR& COLOR::setrgb(short int r, short int g, short int b)
{
	lastoper=(init_color(color, r, g, b)!=ERR);
	return *this;
}

COLOR& COLOR::operator=(short col)
{
	color=col;
	return *this;
}

short COLOR::get()
{
	return color;
}

COLOR& COLOR::getrgb(short int& r, short int& g, short int& b)
{
	lastoper=(color_content(color, &r, &g, &b)!=ERR);
	return *this;
}

#ifdef NCURSES_VERSION

int COLOR::use_default()
{
	int x;
	x=use_default_colors();
	if(x!=ERR)
	{
		uses_default=true;
		return x;
	}
	uses_default=false;
	return x;
}

int COLOR::assume_default(int fg, int bg)
{
	int x;
	x=assume_default_colors(fg,bg);
	if(x!=ERR)
	{
		uses_default=true;
		return x;
	}
	uses_default=false;
	return x;
}

bool COLOR::using_default()
{
	return uses_default;
}

#endif

bool COLOR::can_change()
{
	return can_change_color();
}

bool COLOR::has()
{
	return has_colors();
}

int COLOR::start()
{
	uses_default=false;
	PAIR::nextpair=1;
	return start_color();
}

bool COLOR::good()
{
	return lastoper;
}














// refresh

WINPLUS& WINPLUS::fresh()
{
	lastoper=(wrefresh(W)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::noutfresh()
{
	lastoper=(wnoutrefresh( W )!=ERR);
	return *this;
}




// a-

WINPLUS& WINPLUS::aoff(int attrs)
{
	lastoper=(wattroff(W,attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aon(int attrs)
{
	lastoper=(wattron(W,attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aset(int attrs)
{
	lastoper=(wattrset(W,attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::cset(short color_pair_number, void* opts)
{
	lastoper=(wcolor_set(W,color_pair_number, opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::cset(short color_pair_number)
{
	lastoper=(wcolor_set(W,color_pair_number, NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::astandend()
{
	lastoper=(wstandend(W)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::astandout()
{
	lastoper=(wstandout(W)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aget(attr_t& attrs, short& pair, void *opts)
{
	lastoper=(wattr_get(W, &attrs, &pair, opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aget(attr_t& attrs, short& pair)
{
	lastoper=(wattr_get(W, &attrs, &pair, NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aoff(attr_t attrs, void *opts)
{
	lastoper=(wattr_off(W,attrs,opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aoff(attr_t attrs)
{
	lastoper=(wattr_off(W,attrs,NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aon(attr_t attrs, void *opts)
{
	lastoper=(wattr_on(W,attrs,opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aon(attr_t attrs)
{
	lastoper=(wattr_on(W,attrs,NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aset(attr_t attrs, short pair, void *opts)
{
	lastoper=(wattr_set(W,attrs,pair,opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aset(attr_t attrs, short pair)
{
	lastoper=(wattr_set(W,attrs,pair,NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::chga(int n, attr_t attr, short color, const void *opts)
{
	lastoper=(wchgat(W,n,attr,color,opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::chga(int n, attr_t attr, short color)
{
	lastoper=(wchgat(W,n,attr,color,NULL)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aoff(ATTR attrs)
{
	lastoper=(wattroff(W,attrs.attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aon(ATTR attrs)
{
	lastoper=(wattron(W,attrs.attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aset(ATTR attrs)
{
	lastoper=(wattrset(W,attrs.attrs)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aoff(PAIR attrs)
{
	lastoper=(wattroff(W,COLOR_PAIR(attrs.pair))!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aon(PAIR attrs)
{
	lastoper=(wattron(W,COLOR_PAIR(attrs.pair))!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aset(PAIR attrs)
{
	lastoper=(wattrset(W,attrs.pair)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::aget(attr_t& attrs, PAIR& pair, void *opts)
{
	short p;
	lastoper=(wattr_get(W, &attrs, &p, opts)!=ERR);
	pair.pair=PAIR_NUMBER(p);
	pair_content(p,&(pair.fore.color),&(pair.back.color));
	return *this;
}

WINPLUS& WINPLUS::aset(attr_t attrs, PAIR pair, void *opts)
{
	lastoper=(wattr_set(W, attrs, COLOR_PAIR(pair.pair), opts)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::chga(int n, attr_t attr, COLOR color, const void *opts)
{
	lastoper=(wchgat(W,n,attr,color.color,opts)!=ERR);
	return *this;
}














int WINPLUS::chget()
{
	return wgetch(W);
}




// std::cin & std::cout -like
WINPLUS& WINPLUS::operator<<(int k)
{
	print("%d",k);
	return *this;
}

WINPLUS& WINPLUS::operator<<(float k)
{
	print("%f",k);
	return *this;
}

WINPLUS& WINPLUS::operator<<(char k)
{
	print("%c",k);
	return *this;
}

WINPLUS& WINPLUS::operator<<(double k)
{
	print("%lf",k);
	return *this;
}

WINPLUS& WINPLUS::operator<<(char* k)
{
	print("%s",k);
	return *this;
}

WINPLUS& WINPLUS::operator<<(std::string k)
{
	print("%s",k.c_str());
	return *this;
}




WINPLUS& WINPLUS::operator>>(int& k)
{
	scan("%d", &k);
	return *this;
}

WINPLUS& WINPLUS::operator>>(float& k)
{
	scan("%f",&k);
	return *this;
}

WINPLUS& WINPLUS::operator>>(char& k)
{
	scan("%c",&k);
	return *this;
}

WINPLUS& WINPLUS::operator>>(double& k)
{
	scan("%lf",&k);
	return *this;
}

WINPLUS& WINPLUS::operator>>(char*& k)
{
	scan("%s",k);
	return *this;
}

WINPLUS& WINPLUS::operator>>(std::string& k)
{
	scan("%s",k.c_str());
	return *this;
}






WINPLUS& WINPLUS::move(int y, int x)
{
	wmove(W,y,x);
	return *this;
}



// (v)wprintw & (v)wscanw -like
WINPLUS& WINPLUS::print(char *mask, ...)
{
	int res;
	va_list args;
	va_start(args,mask);
	res=vw_printw(W,mask,args);
	va_end(args);
	lastoper=(res!=ERR);
	return *this;
}

WINPLUS& WINPLUS::vprint(char *mask, va_list args)
{
	lastoper=(vw_printw(W,mask,args)!=ERR);
	return *this;
}

WINPLUS& WINPLUS::scan(char *mask, ...)
{
	int res;
	va_list args;
	va_start(args,mask);
	res=vw_scanw(W,mask,args);
	va_end(args);
	lastoper=(res!=ERR);
	lastscan=res;
	return *this;
}

WINPLUS& WINPLUS::vscan(char *mask, va_list args)
{
	lastoper=(lastscan=vw_scanw(W,mask,args)!=ERR);
	return *this;
}



// class-specific, constructors, initializers, etc.

WINPLUS::WINPLUS(int nlines, int ncols, int begin_y, int begin_x)
{
	W=newwin(nlines,ncols,begin_y,begin_x);
	lastoper=(W!=NULL);
};

WINPLUS::WINPLUS(const WINPLUS& win)
{
	this->W=dupwin(win.W);
	lastoper=(this->W!=NULL);
};

void WINPLUS::destroy()
{
	delwin(W);
}

WINPLUS::~WINPLUS()
{
	WINPLUS::destroy();
}

void WINPLUS::setwpointer(WINDOW *reference)
{
	W=reference;
}

WINDOW *WINPLUS::getwpointer()
{
	return W;
}

bool WINPLUS::good()
{
	return lastoper;
}

int WINPLUS::scanned()
{
	return lastscan;
}

WPLUSMAIN::WPLUSMAIN()
{
	setwpointer(initscr());
	lastoper=(getwpointer()!=NULL);
}

WPLUSMAIN::~WPLUSMAIN()
{
	endwin();
}
